<?php

// translator takes a number of directories and scans all php files in there for
// strings to translate.
//
// then compares to existing translation files and spots obsolete or missing translations.

// require_once('mysession.class.php');
// require_once('translator.class.php');

// $s = new mySession();

        // strange php notice: you must set default timezone explicitly
        // so we get it, ignoring any warnings and then set it as defaulted (or perviously set explicitly)
        $tz = @date_default_timezone_get();
        date_default_timezone_set($tz);

class tlTools {
    static $argv0 = null;
    static $bad = 0;
    static $file = null;
    static $line = null;
    static public function gripe($msg, $thing = null, $text = null) {
	self::output(1, $msg, $thing, $text);
    }
    static public function note($msg, $thing = null, $text = null) {
	self::output(0, $msg, $thing, $text);
    }
    static public function output($incr, $msg, $thing = null, $text = null) {
	global $argv;
	$err = "";
	if (self::$argv0 === null) self::$argv0 = $argv[0];
	$err .= self::$argv0 . ": " . ((self::$file == null) ? "" : (self::$file . ": ")) . ((self::$line == null) ? "" : ("line " . self::$line . ": "));
	if ($thing instanceof tlTableOccurence) {
	    $err .= "{$thing->file}: line {$thing->line}: ";
	} else if ($thing instanceof tlTableTranslation) {
	    $err .= "'{$thing->original}'[{$thing->language}]: ";
	} else if ($thing instanceof tlTableEntry) {
	    $err .= "'{$thing->original}': ";
	} else if (is_int($thing)) {
	    $err .= token_name($thing) . " ($text): ";
	} else if ($thing === null) {
	} else $err .= "??($thing): ";
	$err .= "$msg\n";
	fwrite(STDERR, $err);
	self::$bad += $incr;
    }
    static public function setPos($file = null, $line = null) {
	if ($file != self::$file) self::$bad = 0;
	self::$file = $file;
	self::$line = $line;
    }
    static public function setLine($line) {
	self::$line = $line;
    }

    public function fine() {
	return self::$bad == 0;
    }

    public function strip($string) {
	$c = substr($string, 0, 1);
	if (($c == '"') || ($c == "'")) $string = substr($string, 1);
	$c = substr($string, strlen($string) - 1);
	if (($c == '"') || ($c == "'")) $string = substr($string, 0, strlen($string) - 1);
	return $string;
    }

    public function quote($string) {
	// quote string content so it could be re-printed as php string in '...'
	$q = str_replace("\\", "\\\\", $string);
	$q = str_replace("'", "\\'", $q);
	return $q;
    }

}

class tlTableOccurence {
    // a file / line pair
    var $file;
    var $line;

    public function __construct($file = null, $line = null)  {
	if ($file !== null) $this->file = $file; else $this->file = tlTools::$file;
	if ($line !== null) $this->line = $line; else $this->line = tlTools::$line;
    }

    function __toString() {
	return ($this->file . "/" . $this->line);
    }
}

class tlTableTranslation {
    // a translation of an item
    var $original;
    var $language;
    var $translation;
    var $definition;      // tlTableOccurrence

    public function __construct($original, $language, $translation, tlTableOccurence $definition) {
	$this->original = tlTools::strip($original);
	$this->language = $language;
	$this->translation = tlTools::strip($translation);
	$this->definition = $definition;
    }
}

class tlTableEntry {
    // a translatable item
    var $original;
    var $use = array();   // of tlTableOccurence
    var $translations = array();
                          // indexed array of tlTableTranslation, key is language
    var $hint;	          // optional hint, can be used in src like 'tl::tl( /* max 8 chars */ "original")'
    var $delim;           // string delimiter
    var $tpresent;	  // true if translation is in file but not translated
    var $why = "constructor";

    public function __construct($original, $delim) {
	$this->original = $original;
	$this->delim = $delim;
	$this->tpresent = false;
    }

    public function addUse(tlTableOccurence $where) {
	$this->use[] = $where;
    }

    public function addTranslation(tlTableTranslation $how) {
	// if it is a null translation (src/trans identical), then ignore it
	$this->tpresent = true;
	if ($how->original === $how->translation) {
	    return;
	}
	if (!array_key_exists($how->language, $this->translations)) {
	    // new language
	    $this->translations[$how->language] = $how;
	} else {
	    tlTools::gripe("translation already exists", $how);
	    tlTools::note("place of first translation", $this->translations[$how->language]->definition);
	}
    }
}

class tlTable {
    // $tab is an array of translation entries
    // array (
    //     "translatable string" => // tlTableEntry
    // )
    //
    var $tab = array();
    var $files = array();
    var $languages = array();
    var $contents = array();
    var $new = 0;	// new entries
    var $missing = 0;   // untranslated but used entries
    var $unused = 0;	// unused entries (translated or untranslated)
    var $computed = false;

    function addLanguage($language, $file = null, $module = null) {
	$key = $language ;
	if ($module != "") $key = $module . "_" . $key;
	if (array_key_exists($key, $this->files) &&
            ($this->files[$key] != $file)) {
	    tlTools::gripe("duplicate translation table for class '$module', language '$language'");
	} else {
	    if ($file !== null) $this->files[$key] = $file;
	    // if (!in_array($language, $this->languages)) $this->languages[] = $language;
	    $this->languages[$language] = array();
	}
    }

    function __construct() {
    }

    function getEntry($ostring, $why, $hint = null) {
	// gets the entry for a translatable string
	$string = tlTools::strip($ostring);
	if (!array_key_exists($string, $this->tab)) {
	    $this->tab[$string] = new tlTableEntry($string, $ostring[0]);
	} else {
	}
	$this->tab[$string]->hint = $hint;
	$this->tab[$string]->why .= " " . $why;
	return $string;
    }

    function findlanguagefile($lang) {
	if (isset($this->files["$lang"])) {
	    return $this->files["$lang"];
	}
	// print_r($this->files);
	foreach ($this->files as $tag => $file) {
	    if (substr($tag, strpos($tag, "_") + 1) == $lang) return $file;
	}
	tlTools::gripe("no dictionary for '$lang'");
	return null;
    }

    function compute() {
	// compute the result
	// first sort all entries alphabetically
	ksort($this->tab);
	// search for entries, where just the delimiter has changed
	foreach ($this->tab as $string => &$entry) {
        $unused = false;
        if (count($entry->use) == 0) {
            // unused, perhaps missspelled?
            $unused = true;
		}
        $justDelimChanged = false;
	    foreach($this->tab as &$tmpEntry) {
            $strReplace = "";
            if($entry->delim == "\"") $strReplace = "\\\"";
            else $strReplace = "\\'";
            $entOrg = str_replace($strReplace, $entry->delim, $entry->original);
            $strTmpReplace = "";
            if($tmpEntry->delim == "\"") $strTmpReplace = "\\\"";
            else $strTmpReplace = "\\'";
            $tmpEntOrg = str_replace($strTmpReplace, $tmpEntry->delim, $tmpEntry->original);
            if($unused && $entOrg == $tmpEntOrg && $entry->delim != $tmpEntry->delim) {
                $justDelimChanged = true;
                if($entry->tpresent && !$tmpEntry->tpresent) {
                    $tmpEntry->tpresent = true;
                    $tmpEntry->translations = $entry->translations;
                    foreach ($tmpEntry->translations as $lang1 => &$translation) {
                        $translation->original = str_replace($strReplace, $entry->delim, $translation->original);
                        $translation->translation = str_replace($strReplace, $entry->delim, $translation->translation);
                        $translation->original = str_replace($tmpEntry->delim, $strTmpReplace, $translation->original);
                        $translation->translation = str_replace($tmpEntry->delim, $strTmpReplace, $translation->translation);
                    }
                }
                unset($this->tab[$string]);
                break;
            }
        }
	}
	unset($entry);
	unset($translation);
	foreach ($this->tab as $string => &$entry) {
	    // for each translated item
	    $unused = false;
	    $ambiguous = false;
	    $new = false;
	    $missing = false;
	    if (!$entry->tpresent) { $new = true; $this->new++; }
	    if (count($entry->use) == 0) {
		// unused, perhaps missspelled?
		$unused = true;
		$this->unused++;
		foreach ($entry->translations as $translation) {
                    if (!isset($this->languages[$translation->language]['unused'])) $this->languages[$translation->language]['unused'] = 1;
		     $this->languages[$translation->language]['unused']++;
		}
	    } else if (count($entry->use) > 1) {
		$ambiguous = true;
	    }
	    if (!$unused) {
		if (count($entry->translations) == 0) {
		    // used but not translated, probably new in source
		    $this->missing++;
		    $missing = true;
		}
		// some translations missing
		foreach ($this->languages as $language => &$langattribs) {
		    if (!array_key_exists($language, $entry->translations)) {
                        if (!isset($langattribs['missing'])) $langattribs['missing'] = 1;
                        else $langattribs['missing']++;
		    }
		}
	    }

	    // write out updated translation tables for each language
	    // an existing translation is written out to the file it was found in
	    // a missing translation is written into a file that is named similar to an existing translation
	    // any untranslated item is written to the standard translation file for that language
	    // create lieu of use tag
	    $remarks = array();
	    if ($unused) $remarks[] = "@unused";
	    if ($ambiguous) $remarks[] = "@ambiguous";
	    if ($new) $remarks[] = "@new";
	    $remarks[] = implode(", ", $entry->use);
	    $used = implode("; ", $remarks);
	    $havelanguages = array();
	    // first write all existing translations
            
	    foreach ($entry->translations as $lang1 => &$translation) {
		// remember translated language
		$havelanguages[$lang1] = 1;
		// write out 
                if (!isset($this->contents[$translation->definition->file])) $this->contents[$translation->definition->file] = "";
		$this->contents[$translation->definition->file] .= "    // $used: \r\n";
		if ($entry->hint != null) $this->contents[$translation->definition->file] .= "    $entry->hint\r\n";
		$this->contents[$translation->definition->file] .= "        {$entry->delim}{$translation->original}{$entry->delim} \r\n";
		$this->contents[$translation->definition->file] .= "     => {$entry->delim}{$translation->translation}{$entry->delim},\r\n\r\n";
	    }
	    // now see if languages were missing
	    if (!$unused) foreach ($this->languages as $lang2 => &$langattribs) {
		$key =  $this->findlanguagefile($lang2);
		if (!isset($havelanguages[$lang2]) || $havelanguages[$lang2] !== 1) {
		    if ($key == null) continue;
                    if (!isset($this->contents[$key])) $this->contents[$key] = "";
		    $this->contents[$key] .= "    //     @missing; $used: \r\n";
		    if ($entry->hint != null) $this->contents[$key] .= "    $entry->hint\r\n";
		    $this->contents[$key] .= "          {$entry->delim}{$entry->original}{$entry->delim} \r\n";
		    $this->contents[$key] .= "       => {$entry->delim}{$entry->original}{$entry->delim}, \r\n\r\n";
		} 
	    }
	}
	tlTools::setpos();
	tlTools::note(count($this->tab) . " different strings");
	if ($this->new > 0) tlTools::note("{$this->new} new strings");
	if ($this->missing > 0) tlTools::gripe("{$this->missing} missing translations");
	if ($this->unused > 0) tlTools::note("{$this->unused} unused translations");
	foreach ($this->languages as $lang3 => &$langattribs) {
	    if (isset($langattribs['unused']) && $langattribs['unused'] > 0) tlTools::note("{$langattribs['unused']} unused translations in '{$lang3}'");
	    if (isset($langattribs['missing']) && $langattribs['missing'] > 0) tlTools::gripe("{$langattribs['missing']} untranslated strings in '{$lang3}'");
	}
	global $argv;
	foreach ($this->contents as $fn => &$body) {
	    $nfn = basename($fn, ".inc") . ".txt";
	    $lang = "oops??"; foreach ($this->files as $l => $f) if ($f == $fn) { $lang = $l; break; }
	    $b = "<?php\r\n";
            $b .= "    //\r\n";
            $b .= "    // $nfn for language '$lang'\r\n";
            $b .= "    // updated by {$argv[0]} at " . date("r") . "\r\n";
            $b .= "    //\r\n\r\n";
            $b .= "    // to create a new translation change the right hand side of the pairs in the array below\r\n";
            $b .= "\r\n";
	    $b .= "    // an entry marked as '@missing' needs to be translated: the next 2 lines contain \r\n";
	    $b .= "    //                                                      the original string and the syntax already\r\n";
	    $b .= "    // an entry starting with '@xyz@...' must retain this tag also in the translation.  This tag will not\r\n";
	    $b .= "    //                                                      show up in the user interface and serves to \r\n";
	    $b .= "    //                                                      distinguish different translations for a single string\r\n";
            $b .= "\r\n";
	    $b .= "    // If a translation is identical to the original (e.g. 'Name' in English to 'Name' in German), tag the translation with '@@':\r\n";
	    $b .= "    //                 'Name' => '@@Name'                   \r\n";
	    $b .= "    //                                                      to mark it as translated\r\n";
            $b .= "\r\n";
            $b .= "function _translator_$lang(&\$target) {\r\n";
            $b .= "\r\n";
            $b .= "    \$target = array (\r\n";
            $b .= "$body\r\n";
            $b .= "    );\r\n";
            $b .= "}\r\n";
            $b .= "?>\r\n";
	    $body = $b;
	}
	$this->computed = true;
	return tlTools::fine();
    }

    function dump() {
	// write out updated dictionary files
	tlTools::setPos();
	if (!$this->computed) $this->compute();
	foreach ($this->contents as $fn => &$b) {
	    $nfn = dirname($fn) . "/" . basename($fn, ".inc") . ".txt";
	    tlTools::setPos($nfn);
	    if (file_put_contents ($nfn, $b) === false) {
		tlTools::gripe("failed to write");
	    } else {
		tlTools::note(strlen($b) . " bytes written");
	    }
	}
	return tlTools::fine();
    }
}



class tlTransParser {

    const stIdle = 0,
	  stFunction = 1, 
	  stName = 2,
	  stArg = 3,
	  stOpen = 4,
	  stLeft = 5,
	  stRight = 6,
	  stClose = 7,
          stArrow = 8,
	  stEnd = 999;

    private function scanTranslation(tlTable &$tab, $file, &$tokens) {
	$state = self::stIdle;
	$rem = preg_match('/^lang(-([^-]*))?-(.*).inc$/', basename($file), $match);
	if ($rem != 1) {
	    tlTools::gripe("this is not a valid file name for a translation db");
	    return false;
	}
	$function = str_replace("-", "_", "_translator{$match[1]}-{$match[3]}");
	$language = $match[3];
	$module = $match[2];
	$tab->addLanguage($language, $file, $module);

	foreach ($tokens as $t) {
	    if (is_string($t)) continue;
	    list($id, $text, $line) = $t;
	    tlTools::setLine($line);
	    if ($id == T_WHITESPACE) continue;
	    if ($id == T_COMMENT) continue;
	    // print "transparser: $state: token($text)=" . token_name($id) . "\n";
	    switch($state) {
	    case self::stIdle : 
		if ($id == T_FUNCTION) $state = self::stFunction;
		break;
	    case self::stFunction :
		if (($id == T_STRING) && ($text == $function)) {
		    $state = self::stName;
		} else {
		    tlTools::gripe("unexpected function name", $id, $text);
		    $state = self::stIdle;
		}
		break;
	    case self::stName : 
		if ($id == T_VARIABLE) $state = self::stArg;
		else $state = self::stIdle;
		break;
	    case self::stArg :
		if ($id == T_VARIABLE) $state = self::stOpen;
		else $state = self::stIdle;
		break;
	    case self::stOpen :
		if ($id == T_ARRAY) {
		    $state = self::stLeft;
		}
		else $state = self::stIdle;
		break;
	    case self::stLeft :
		if ($id == T_CONSTANT_ENCAPSED_STRING) {
		    $state = self::stArrow;
		    $left = $text;
		}
		else if ($id == T_ENCAPSED_AND_WHITESPACE) {
		    tlTools::gripe("do not use variables in left side of translations", $id, $text);
		    $state = self::stIdle;
		}
		else $state = self::stIdle;
		break;
	    case self::stArrow : 
		if ($id == T_DOUBLE_ARROW) $state = self::stRight;
		else $state = self::stIdle;
		break;
	    case self::stRight : 
		if ($id == T_CONSTANT_ENCAPSED_STRING) {
		    $state = self::stLeft;
		    $k = $tab->getEntry($left, "transl.");
		    $tab->tab[$k]->addTranslation(new tlTableTranslation($left, $language, $text, new tlTableOccurence()));
		}
		else if ($id == T_ENCAPSED_AND_WHITESPACE) {
		    tlTools::gripe("do not use variables in right side of translations", $id, $text);
		    $state = self::stIdle;
		}
		else $state = self::stIdle;
		break;
	    }
	}
	return tlTools::fine();
    }


    function translate(&$tab, $file) {
	tlTools::setPos($file);
	$source = @file_get_contents($file);
	if ($source === false) return(false);
	tlTools::setPos($file, 0);
	return ($this->scanTranslation($tab, $file, token_get_all($source)));
    }

}

class tlSourceParser {
    const stIdle = 0,
	  stScope = 1, 
	  stColon = 2,
	  stFunction = 3,
	  stOpen = 4,
	  stArgument = 5,
	  stClose = 6,
	  st__end = 99;
    const ALWAYS = 999999;
    var $skip = self::ALWAYS;
    
    function __construct($warn = false) {
	$this->skip = $warn ? 0 : self::ALWAYS;
    }

    private function scanSource(&$tab, &$tokens) {
	$state = self::stIdle;
	$skip = $this->skip;
	$ignore = array();
	$untranslatable = 0;
	foreach ($tokens as $t) {
	    if (is_string($t)) continue;
	    list($id, $text, $line) = $t;
	    tlTools::setLine($line);
	    if ($state == self::stIdle) $func = null;

	    if ($id == T_WHITESPACE) continue;
	    if ($id == T_COMMENT) {
		// meta commands?
		if (preg_match('@((/\*)|(//))[\t ]*([Tt]-([a-zA-Z0-9-]+)(\(([0-9a-zA-Z_]+)\))?)@', $text, $match) > 0) {
		    $cmd = strtolower($match[5]);
		    $arg = isset($match[7]) ? $match[7] : "";
		    switch ($cmd) {
		    case 'skip' :
			if ($arg != "") {
			    if ($arg >= $skip) $skip = $arg;
			    break;
			} /* else fall thru */
		    case 'off' :
			$skip = self::ALWAYS; break;
		    case 'on' :
			$skip = $this->skip; break;
		    default :
			tlTools::gripe("unknown command: $cmd");
		    }
		}
		if ($state != self::stOpen) continue;
	    }

	    // print "sourceparser: $state: token($text)=" . token_name($id) . "\n";
	    switch($state) {
	    case self::stIdle : 
		$hint = null;
		if (($id == T_STRING) && ($text == "tl")) $state = self::stScope;
		switch ($id) {
		case T_ENCAPSED_AND_WHITESPACE :
		case T_CONSTANT_ENCAPSED_STRING :
		    if ($skip > 0)  {
			// skip this string
			$skip--;
		    } else {
			if (!array_key_exists($text, $ignore)) {
			    tlTools::note("untranslatable string: $text");
			    $untranslatable++;
			    $ignore[$text] = 1;
			} 
		    }
		}
		break;
	    case self::stScope :
		if ($id == T_DOUBLE_COLON) $state = self::stFunction;
		else $state = self::stIdle;
		break;
	    case self::stFunction :
		if (($id == T_STRING) &&
		    (($text == "tl") || ($text == "tlx"))) {
		    $state = self::stOpen;
		    $func = $text;
		} else $state = self::stIdle;
		break;
	    case self::stOpen : 
		if ($id == T_CONSTANT_ENCAPSED_STRING) {
		    $k = $tab->getEntry($text, "use", $hint);
		    $tab->tab[$k]->addUse(new tlTableOccurence());
		}
		else if ($id == T_ENCAPSED_AND_WHITESPACE) tlTools::gripe("must not use variables in translatable string", $id, $text);
		else if ($id == T_COMMENT) { $hint = $text; continue; }
		else tlTools::gripe("use '...' instead of " . token_name($id) . " for translatable string", $id, $text);
		$state = self::stIdle;
	    }
	}
	if ($untranslatable > 0) tlTools::note("$untranslatable untranslatable strings");
	return tlTools::fine();
    }


    function translate(&$tab, $file) {
	tlTools::setPos($file);
	$source = @file_get_contents($file);
	if ($source === false) return(false);
	tlTools::setPos($file, 0);
	return ($this->scanSource($tab, token_get_all($source)));
    }
}


if (php_sapi_name() == "cli") {
    /*
     * call
     * translator [ -d <dictionary> ] [ -D <d1> <d2> <...> -D ] files
     * [ more than one -d option allowed
     */

    // print_r($argv);
    $tab = new tlTable;
    $exit = 0;
    $ndicts = 0;
    $warn = false;
    for ($i = 1; $i < count($argv); $i++) {

	tlTools::setPos();
	if ($argv[$i] == "-d") {
	    $ndicts = 1;
	    continue;
	}
	if ($argv[$i] == "-D") {
	    if ($ndicts > 0) $ndicts = 0;
	    else $ndicts = 9999;
	    continue;
	} 
	if ($argv[$i] == "-w") $warn = true;
	if ($ndicts > 0) {
	    $ft = new tlTransParser();
	    $type = "dictionary";
	    $ndicts--;
	} else {
	    $ft = new tlSourceParser($warn);
	    $type = "source file";
	}
	tlTools::setPos($argv[$i]);
	tlTools::note("parsing as {$type}");
	tlTools::setPos();
	if (!$ft->translate($tab, $argv[$i])) {
	    tlTools::gripe("failed to parse {$argv[$i]}");
	    $exit++;
	}
    }

    tlTools::setPos();
    if (!$tab->dump()) {
	tlTools::setPos();
	tlTools::gripe("failed to update some or all of your dictionaries");
	$exit++;
    }

    // print_r($tab->files);
    // print_r($tab->languages);
    // print_r($tab->contents);
    // print_r($tab);
    // exit($exit);
} 
?>
